# pages module defines classes for interacting with wikipedia pages.
# There are separate classes defined for each page with their own
# defined methods for performing certain actions.

from . import browser
from . import util
from .assets.languages import LANGUAGES

import logging
import re
import selenium
import time

PAGE_BASE_URL = 'https://www.wikipedia.org/'

class PageRootObject:
    """
    Common interface methods for working with pages. The specific
    page classes below inherit these methods and define additional methods
    so every page has available these methods and any additional
    methods they define.

    In here are some re-used methods to click links and highlight
    elements in the browser.
    """
    def __init__(self, driver):
        """
        Object constructor for initializing the instance of this
        class with internal variables needed.

        Args:
            driver: Reference to the selenium webdriver object
        that is used to interface with the web browser.
        """
        self.driver = driver

    def click(self, el): 
        """
        Clicks a link in the browser and also highlights it to the
        end user.

        Args:
            el: selenium element to be clicked. Typically an anchor
        html link in the webpage.
        """
        self.highlight(el, 'red')
        DELAY=0
        time.sleep(DELAY)
        util.breakpoint()
        el.click()

    def highlight(self, el, color):
        """
        Highlights an html element in the web browser by changing the
        background color as well as making the text bold.

        The implementation uses javascript to alter the css of the element.

        Args:
            el: selenium element to be highlighted.
            color: background color to highlight. Input can be one of
        'red', 'blue', or hex code such as '#ffffff'.
        """
        # Note: The way hex codes work is there is 1 byte (2 hex characters)
        # for every color. #RRGGBB for (red, green, blue). This can be thought
        # of as an integer 0-255 for red, green, and blue in base-16 hexadecimal.
        # For example, #ff0000 is bright red while #002f00 is light green
        # and #ffff00 is full yellow.
        if color == 'red':
            color = '#ff9292'
            # js = 'arguments[0].setAttribute("style", "background: %s;font-weight:bold;color:#000")' % '#ff9292'
        elif color == 'blue':
            color = '#9292ff'
            # js = 'arguments[0].setAttribute("style", "background: %s;font-weight:bold;color:#000")' % '#9292ff'
        else:
            # color = color
            pass
        js = 'arguments[0].setAttribute("style", "background: %s;font-weight:bold;color:#000")' % color
        self.driver.execute_script(js, el)

# Note: This is the syntax for class inheritance. LandingPage is a new type of object that inherits
# everything from the PageRootObject type. With this, you can call LandingPage.highlight() which
# is a method defined in PageRootObject.
class LandingPage(PageRootObject):
    """
    Interface for working with the wikipedia.org landing page. This page has links to
    select a language and go to the respective wikipedia root page.
    """

    # Note: This is the LandingPage() class constructor. The constructor is a method
    # that is executed when a new object of this class is created. All it does right now is
    # reference the parent (PageRootObject) constructor method and call it. This
    # calls PageRootObject.__init__(driver) which then makes the web driver available
    # in the object instance.
    def __init__(self, driver=None):
        super().__init__(driver)

    def goto_landing_page(self):
        """
        Navigates the browser to www.wikipedia.org
        """
        self.driver.get(PAGE_BASE_URL)

    def select_language(self, language):
        lang_text = LANGUAGES.get(language)
        try:
            link = self.driver.find_element_by_partial_link_text(lang_text)
            self.click(link)
            return True
        except selenium.common.exceptions.NoSuchElementException as ex:
            logging.warn('failed to find language: %s as %s' % (language, lang_text))
            return False

class MainPage(PageRootObject):
    """
    Interface for a selected language root page. This has the link to go to a random article
    and has a featured article. An example url for this is https://en.wikipedia.org.
    """
    def __init__(self, driver=None):
        super().__init__(driver)

    def goto_random_article(self):
        link = self.driver.find_element_by_xpath('//li[contains(@id, "n-randompage")]/a')
        self.click(link)

class ArticlePage(PageRootObject):
    """
    Interface for a wikipedia article page. Here are defined some utility methods to
    try and click the first valid link and extract some information from the page.
    """

    # Here are static class-scoped variables that are needed to work with the page.
    # These are used to locate html elements in the web browser. There are many
    # ways to locate elements but one of the best if available is locating by id. It's
    # not enforced but the html specification mandates that element id's are unique
    # so if you can select by id in a semantically correct web page, you can correctly
    # select unique elements with high confidence.
    elements = {
        'main-window-content-text-id': 'mw-content-text',
        'article-title-id': 'firstHeading',
    }

    def __init__(self, driver=None):
        super().__init__(driver)

    def get_title(self):
        """
        Returns the article title.
        """
        heading = self.driver.find_element_by_id(ArticlePage.elements['article-title-id'])
        return heading.text

    def click_first_link(self):
        """
        Attempts to click the first valid link in the article. Some work is
        done to skip over certain links but the implementation breaks in some
        edge cases. It's close but not perfect for every article text.
        """
        return self._iterate_paragraphs()

    # Note: Here this method has it's name prepended with a single underscore.
    # This is a convention that communicates to the developer that these methods
    # are internal private methods. That means they are not meant to be exposed
    # to the external interface. Python does not restrict calling these methods.
    # You can still call ArticlePage._iterate_paragraphs() but the prefix
    # underscore tells you that it is not intended to be exposed and may be
    # unsafe to call. Depending on the implementation, it may not make sense
    # to directly call this method and may result in undefined and unexpected
    # behavior. _iterate_paragraphs is called internally from the exposed
    # click_first_link() but is never invoked externally.
    def _iterate_paragraphs(self):
        """
        Iterates through paragraphs in the page and attempts to find the first
        valid link. Sometimes the first paragraph does not have a link so this
        needs to go through a few paragraphs and it does not make sense to
        operate on the entire article every time when we're just looking for
        the first link, for performance optimization.
        """
        main_window = self.driver.find_element_by_id(ArticlePage.elements['main-window-content-text-id'])

        # Note on xpath (more advanced web automation tool):
        # xpath is another language that allows you to concisely specify an
        # element in the page (or more generally any xml or html structured
        # text). You can define a hierarchical tree structure that an element
        # must have, attributes that any of the nodes must have or not have,
        # and even some more complex functionality. For example, you can
        # say in xpath to give me all the links that are a child of the
        # navigation menu. This xpath here looks for paragraph elements
        # that fall under this structure:
        #
        # <div id="mw-content-text">
        #   <div class="mw-parser-output">
        #     <p>
        #       ...wikipedia article content...
        #
        #       and does NOT have:
        #       <span><span id="coordinates">...</span></span>
        #
        #     </p>
        #   </div>
        # </div>
        xpath_str = './div[contains(@class, "mw-parser-output")]/p[not(span/span[contains(@id, "coordinates")])]'
        paragraphs = main_window.find_elements_by_xpath(xpath_str)
        for p in paragraphs:
            # Return code indicates the success status of _parse_paragraph().
            # In this case, an rc of True means that it was able to find a
            # link and we stop going through paragraphs.
            rc = self._parse_paragraph(p)
            if rc:
                return True

    def _parse_paragraph(self, p):
        """
        Attempts to find a valid link in the paragraph element sent in.

        Args:
            p: Reference to selenium paragraph element. This is a paragraph
        taken from the article.
        """
        links = p.find_elements_by_xpath('.//a')
        if len(links) == 0:
            return False
        for link in links:
            logging.getLogger('pages').debug('processing link: %s' % link.text)
            if not self._is_valid_link(p, link):
                logging.getLogger('pages').debug('skipping link inside parenthesis: %s' % link.text)
                self.highlight(link, 'blue')
                continue
            self.highlight(link, 'red')
            logging.getLogger('pages').info('selected link: %s' % link.text)
            self.click(link)
            return True

    def _is_valid_link(self, p, el):
        """
        Returns if the implementation decides to skip this link. You can
        see the reasons we invalidate and skip a link here. If it's
        inside parenthesis, is a footnote, is a pronounciation guide or
        audio link, we choose to skip it.
        """
        if self._is_link_in_parenthesis(p, el):
            return False
        if self._is_link_a_footnote(el):
            return False
        if self._is_link_pronounciation(el):
            return False
        if self._is_link_audio(el):
            return False
        if self._is_not_wikipedia(el):
            return False
        return True

    def _is_link_in_parenthesis(self, p, el):
        """
        Determine if a given link element is inside a set
        of textual parenthesis.
        """
        # Implementation notes (mg):
        # I've tried a few different ways to do this and it's
        # hard to get it to work in every case. I want to avoid
        # certain links and usually avoid links inside parenthetical
        # notes. Some edge cases are nested parenthesis, links with
        # non-english characters (which are displayed with a tree
        # of elements in the html rather than a simple link). And
        # sometimes, the link inside the parenthesis may be a valid
        # target. I've made it so that skipped links show up as blue
        # and determined-valid links highlight as red.
        link_text = el.get_attribute('outerHTML')
        p_text = p.get_attribute('innerHTML')

        # Note on regular expressions (advanced):
        # Regular expressions or regexes are another language for
        # matching patterns in raw text (regex is blind to html structure).
        # Regular expressions are notorious because they can be
        # hard to understand and read but they are extremely
        # expressive, i.e. you can convey a great deal of functionality
        # in one line of code. This one below is among the simplest and
        # just grabs text inside a pair of parenthesis.
        regex_str = '\(.*?\)'
        regex = re.compile(regex_str, flags=re.UNICODE)
        match = regex.search(p_text)
        if not match:
            # There are no parenthesis at all in this paragraph.
            return False

        while match is not None:
            # There may be multiple parenthesis (or nested). This
            # iterates through them and checks if the links html
            # is present inside these parenthesis.
            #
            # Care must be taken with regular expressions as they are
            # user/developer unfriendly, hard-to-read, and unforgiving.
            # For example, what happens when you try to extract text inside parenthesis
            # inside of "(some words) some more words (even more words)", you
            # can match unpaired parenthesis. This means "some more words" is indeed
            # inside a pair of parenthesis, the pair given by the entire string.
            # The code is quite dumb and does exactly what you tell it to.
            # Often this leads to unexpected results because the computer
            # is so literal.
            match_text = match.group(0)
            match_idx = match.end(0)
            if link_text in match_text:
                return True
            match = regex.search(p_text, match_idx+1)

        return False

        # Side Teaser: This is a regular expression to match valid email
        # addresses. It reads 1-or more alphanumeric characters (also _ and . and -)
        # followed by a '@' followed by 1-or more alphanumeric characters
        # (also . and -), followed by a period and between 2 and 6 lower-cased
        # characters a-z (and . accepted).
        # 
        # /^([a-z0-9_\.-]+)@([\da-z\.-]+)\.([a-z\.]{2,6})$/
        #
        # You can see how they get messy quick but one line of regex could
        # do the same thing that would take 100 lines of imperative code.

    def _is_link_a_footnote(self, el):
        # Some links are anchors to footnotes, e.g. [1] that points to a source
        # at the bottom of the page. These aren't valid links for our purpose
        # so this method looks for that and determines if the reference element
        # appears to be a link to a footnote.
        href = el.get_attribute('href')
        if '#cite_note' in href:
            return True
        if '#cnote' in href:
            return True
        return False

    def _is_link_pronounciation(self, el):
        # Some links point to the wikipedia IPA (international phonetic
        # alphabet) pronounciation help page. We don't want to click these
        # links so we scan for and ignore them.
        href = el.get_attribute('href')
        if '/wiki/Help:IPA' in href:
            return True
        return False

    def _is_link_audio(self, el):
        # Some links are audio playback pronounciations. We look for these
        # by checking for the file extension .ogg (an audio file format,
        # ogg-vorbis) and ignoring links if they are of that type.
        href = el.get_attribute('href')
        if '.ogg' in href:
            return True
        return False 

    def _is_not_wikipedia(self, el):
        # Some links point to websites outside of wikipedia, we skip those
        # with this.
        href = el.get_attribute('href')
        if 'wikipedia.org' not in href:
            return True
        return False 

